from ..print import info as info, warning as warning
from .main import install_program as install_program

def download_executable(download_urls: dict[str, str], program_name: str, append_to_path: str = '') -> bool:
    """ Ask the user if they want to download the program (ex: waifu2x-ncnn-vulkan).
\tIf yes, try to download the program from the GitHub releases page.

\tArgs:
\t\tdownload_urls  (dict[str, str]):  The URLs to download the program from.
\t\tprogram_name   (str):             The name of the program to download.

\tReturns:
\t\tbool: True if the program is now ready to use, False otherwise.
\t"""
def check_executable(executable: str, executable_help_text: str, download_urls: dict[str, str], append_to_path: str = '') -> None:
    ''' Check if the executable exists, optionally download it if it doesn\'t.

\tArgs:
\t\texecutable            (str):             The path to the executable.
\t\texecutable_help_text  (str):             The help text to check for in the executable\'s output.
\t\tdownload_urls         (dict[str, str]):  The URLs to download the executable from.
\t\tappend_to_path        (str):             The path to append to the executable\'s path.
\t\t\t(ex: "bin" if executables are in the bin folder)
\t'''



from collections.abc import Callable as Callable, Iterable, Iterator
from typing import Any, IO, TextIO, TypeVar

RESET: str
RED: str
GREEN: str
YELLOW: str
BLUE: str
MAGENTA: str
CYAN: str
LINE_UP: str
BAR_FORMAT: str
T = TypeVar('T')
previous_args_kwards: tuple[Any, Any]
nb_values: int
import_time: float

def colored_for_loop(iterable: Iterable[T], desc: str = 'Processing', color: str = ..., bar_format: str = ..., ascii: bool = False, **kwargs: Any) -> Iterator[T]:
    ''' Function to iterate over a list with a colored TQDM progress bar like the other functions in this module.

\tArgs:
\t\titerable\t(Iterable):\t\t\tList to iterate over
\t\tdesc\t\t(str):\t\t\t\tDescription of the function execution displayed in the progress bar
\t\tcolor\t\t(str):\t\t\t\tColor of the progress bar (Defaults to MAGENTA)
\t\tbar_format\t(str):\t\t\t\tFormat of the progress bar (Defaults to BAR_FORMAT)
\t\tascii\t\t(bool):\t\t\t\tWhether to use ASCII or Unicode characters for the progress bar (Defaults to False)
\t\tverbose\t\t(int):\t\t\t\tLevel of verbosity, decrease by 1 for each depth (Defaults to 1)
\t\t**kwargs:\t\t\t\t\t\tAdditional arguments to pass to the TQDM progress bar

\tYields:
\t\tT:\t\tEach item of the iterable

\tExamples:
\t\t>>> for i in colored_for_loop(range(10), desc="Time sleeping loop"):
\t\t...     time.sleep(0.01)
\t\t>>> # Time sleeping loop: 100%|██████████████████| 10/10 [ 95.72it/s, 00:00<00:00]
\t'''
def info(*values: Any, color: str = ..., text: str = 'INFO ', prefix: str = '', file: TextIO | list[TextIO] | None = None, **print_kwargs: Any) -> None:
    ''' Print an information message looking like "[INFO HH:MM:SS] message" in green by default.

\tArgs:
\t\tvalues\t\t\t(Any):\t\t\t\t\tValues to print (like the print function)
\t\tcolor\t\t\t(str):\t\t\t\t\tColor of the message (default: GREEN)
\t\ttext\t\t\t(str):\t\t\t\t\tText of the message (default: "INFO ")
\t\tprefix\t\t\t(str):\t\t\t\t\tPrefix to add to the values
\t\tfile\t\t\t(TextIO|list[TextIO]):\tFile(s) to write the message to (default: sys.stdout)
\t\tprint_kwargs\t(dict):\t\t\t\t\tKeyword arguments to pass to the print function
\t'''
def debug(*values: Any, **print_kwargs: Any) -> None:
    ''' Print a debug message looking like "[DEBUG HH:MM:SS] message" in cyan by default. '''
def alt_debug(*values: Any, **print_kwargs: Any) -> None:
    ''' Print a debug message looking like "[DEBUG HH:MM:SS] message" in blue by default. '''
def suggestion(*values: Any, **print_kwargs: Any) -> None:
    ''' Print a suggestion message looking like "[SUGGESTION HH:MM:SS] message" in cyan by default. '''
def progress(*values: Any, **print_kwargs: Any) -> None:
    ''' Print a progress message looking like "[PROGRESS HH:MM:SS] message" in magenta by default. '''
def warning(*values: Any, **print_kwargs: Any) -> None:
    ''' Print a warning message looking like "[WARNING HH:MM:SS] message" in yellow by default and in sys.stderr. '''
def error(*values: Any, exit: bool = False, **print_kwargs: Any) -> None:
    """ Print an error message (in sys.stderr and in red by default)
\tand optionally ask the user to continue or stop the program.

\tArgs:
\t\tvalues\t\t\t(Any):\t\tValues to print (like the print function)
\t\texit\t\t\t(bool):\t\tWhether to ask the user to continue or stop the program,
\t\t\tfalse to ignore the error automatically and continue
\t\tprint_kwargs\t(dict):\t\tKeyword arguments to pass to the print function
\t"""
def whatisit(*values: Any, print_function: Callable[..., None] = ..., max_length: int = 250, color: str = ..., **print_kwargs: Any) -> None:
    ''' Print the type of each value and the value itself, with its id and length/shape.

\tThe output format is: "type, <id id_number>:\t(length/shape) value"

\tArgs:
\t\tvalues\t\t\t(Any):\t\tValues to print
\t\tprint_function\t(Callable):\tFunction to use to print the values (default: debug())
\t\tmax_length\t\t(int):\t\tMaximum length of the value string to print (default: 250)
\t\tcolor\t\t\t(str):\t\tColor of the message (default: CYAN)
\t\tprint_kwargs\t(dict):\t\tKeyword arguments to pass to the print function
\t'''
def breakpoint(*values: Any, print_function: Callable[..., None] = ..., **print_kwargs: Any) -> None:
    """ Breakpoint function, pause the program and print the values.

\tArgs:
\t\tvalues\t\t\t(Any):\t\tValues to print
\t\tprint_function\t(Callable):\tFunction to use to print the values (default: warning())
\t\tprint_kwargs\t(dict):\t\tKeyword arguments to pass to the print function
\t"""

class TeeMultiOutput:
    ''' File-like object that duplicates output to multiple file-like objects.

\tArgs:
\t\t*files         (IO[Any]):  One or more file-like objects that have write and flush methods
\t\tstrip_colors   (bool):     Strip ANSI color codes from output sent to non-stdout/stderr files
\t\tascii_only     (bool):     Replace non-ASCII characters with their ASCII equivalents for non-stdout/stderr files
\t\tignore_lineup  (bool):     Ignore lines containing LINE_UP escape sequence in non-terminal outputs

\tExamples:
\t\t>>> f = open("logfile.txt", "w")
\t\t>>> sys.stdout = TeeMultiOutput(sys.stdout, f)
\t\t>>> print("Hello World")  # Output goes to both console and file
\t\tHello World
\t\t>>> f.close()\t# TeeMultiOutput will handle any future writes to closed files gracefully
\t'''
    files: tuple[IO[Any], ...]
    strip_colors: bool
    ascii_only: bool
    ignore_lineup: bool
    def __init__(self, *files: IO[Any], strip_colors: bool = True, ascii_only: bool = True, ignore_lineup: bool = True) -> None: ...
    @property
    def encoding(self) -> str:
        ''' Get the encoding of the first file, or "utf-8" as fallback.

\t\tReturns:
\t\t\tstr: The encoding, ex: "utf-8", "ascii", "latin1", etc.
\t\t'''
    def write(self, obj: str) -> int:
        """ Write the object to all files while stripping colors if needed.

\t\tArgs:
\t\t\tobj (str): String to write
\t\tReturns:
\t\t\tint: Number of characters written to the first file
\t\t"""
    def flush(self) -> None:
        """ Flush all files. """
    def fileno(self) -> int:
        """ Return the file descriptor of the first file. """

def remove_colors(text: str) -> str:
    """ Remove the colors from a text """
def is_same_print(*args: Any, **kwargs: Any) -> bool:
    """ Checks if the current print call is the same as the previous one. """
def current_time() -> str:
    ''' Get the current time as "HH:MM:SS" if less than 24 hours since import, else "YYYY-MM-DD HH:MM:SS" '''
def show_version(main_package: str = 'stouputils', primary_color: str = ..., secondary_color: str = ...):
    ''' Print the version of the main package and its dependencies.

\tUsed by the "stouputils --version" command.

\tArgs:
\t\tmain_package\t\t\t(str):\t\t\tName of the main package to show version for
\t\tprimary_color\t\t\t(str):\t\t\tColor to use for the primary package name
\t\tsecondary_color\t\t\t(str):\t\t\tColor to use for the secondary package names
\t'''




# Imports
import itertools

from .keras.all import (
    VGG16,
    VGG19,
    ConvNeXtBase,
    ConvNeXtLarge,
    ConvNeXtSmall,
    ConvNeXtTiny,
    ConvNeXtXLarge,
    DenseNet121,
    DenseNet169,
    DenseNet201,
    EfficientNetB0,
    EfficientNetV2B0,
    EfficientNetV2L,
    EfficientNetV2M,
    EfficientNetV2S,
    MobileNet,
    MobileNetV2,
    MobileNetV3Large,
    MobileNetV3Small,
    ResNet50V2,
    ResNet101V2,
    ResNet152V2,
    SqueezeNet,
    Xception,
)
from .model_interface import ModelInterface

# Other models
from .sandbox import Sandbox


# Create a custom dictionary class to allow for documentation
class ModelClassMap(dict[type[ModelInterface], tuple[str, ...]]):
    pass

# Routine map
CLASS_MAP: ModelClassMap = ModelClassMap({
    SqueezeNet:         ("squeezenet", "squeezenets", "all", "often"),

    DenseNet121:        ("densenet121", "densenets", "all", "often", "good"),
    DenseNet169:        ("densenet169", "densenets", "all", "often", "good"),
    DenseNet201:        ("densenet201", "densenets", "all", "often", "good"),

    EfficientNetB0:     ("efficientnetb0", "efficientnets", "all"),
    EfficientNetV2B0:   ("efficientnetv2b0", "efficientnets", "all"),
    EfficientNetV2S:    ("efficientnetv2s", "efficientnets", "all", "often"),
    EfficientNetV2M:    ("efficientnetv2m", "efficientnets", "all", "often"),
    EfficientNetV2L:    ("efficientnetv2l", "efficientnets", "all", "often"),

    ConvNeXtTiny:       ("convnexttiny", "convnexts", "all", "often", "good"),
    ConvNeXtSmall:      ("convnextsmall", "convnexts", "all", "often"),
    ConvNeXtBase:       ("convnextbase", "convnexts", "all", "often", "good"),
    ConvNeXtLarge:      ("convnextlarge", "convnexts", "all", "often"),
    ConvNeXtXLarge:     ("convnextxlarge", "convnexts", "all", "often", "good"),

    VGG16:              ("vgg16", "vggs", "all"),
    VGG19:              ("vgg19", "vggs", "all"),

    MobileNet:          ("mobilenet", "mobilenets", "all"),
    MobileNetV2:        ("mobilenetv2", "mobilenets", "all", "often"),
    MobileNetV3Small:   ("mobilenetv3small", "mobilenets", "all", "often"),
    MobileNetV3Large:   ("mobilenetv3large", "mobilenets", "all", "often", "good"),

    ResNet50V2:         ("resnet50v2", "resnetsv2", "resnets", "all", "often"),
    ResNet101V2:        ("resnet101v2", "resnetsv2", "resnets", "all", "often"),
    ResNet152V2:        ("resnet152v2", "resnetsv2", "resnets", "all", "often"),

    Xception:           ("xception", "xceptions", "all", "often"),
    Sandbox:            ("sandbox",),
})

# All models names and aliases
ALL_MODELS: list[str] = sorted(set(itertools.chain.from_iterable(v for v in CLASS_MAP.values())))
""" All models names and aliases found in the `CLASS_MAP` dictionary. """

# Additional docstring
new_docstring: str = "\n\n" + "\n".join(f"- {k.__name__}: {v}" for k, v in CLASS_MAP.items())
ModelClassMap.__doc__ = "Dictionary mapping class to their names and aliases. " + new_docstring
CLASS_MAP.__doc__ = ModelClassMap.__doc__




""" Base implementation for machine learning models with common functionality.
Provides shared infrastructure for model training, evaluation, and MLflow integration.

Implements comprehensive workflow methods and features:

Core Training & Evaluation:
- Full training/evaluation pipeline (routine_full)
- K-fold cross-validation with stratified splitting
- Transfer learning weight management (ImageNet, custom datasets)
- Model prediction and evaluation with comprehensive metrics

Hyperparameter Optimization:
- Learning Rate Finder with automatic best LR detection
- Unfreeze Percentage Finder for fine-tuning optimization
- Class weight balancing for imbalanced datasets
- Learning rate warmup and scheduling (ReduceLROnPlateau)

Advanced Training Features:
- Early stopping with configurable patience
- Model checkpointing with delay options
- Additional training data integration (bypasses CV splitting)
- Multi-processing support for memory management
- Automatic retry mechanisms with error handling

MLflow Integration:
- Complete experiment tracking and logging
- Parameter logging (training, optimizer, callback parameters)
- Metric logging with averages and standard deviations
- Model artifact saving and versioning
- Training history visualization and plotting

Model Architecture Support:
- Keras/TensorFlow and PyTorch compatibility
- Automatic layer counting and fine-tuning
- Configurable unfreeze percentages for transfer learning
- Memory leak prevention with subprocess training

Evaluation & Visualization:
- ROC and PR curve generation
- Comprehensive metric calculation (Sensitivity, Specificity, AUC, etc.)
- Training history plotting and analysis
- Saliency maps and GradCAM visualization (single sample)
- Cross-validation results aggregation

Configuration & Utilities:
- Extensive parameter override system
- Verbosity control throughout pipeline
- Temporary directory management for artifacts
- Garbage collection and memory optimization
- Error logging and handling with retry mechanisms
"""
# pyright: reportUnknownMemberType=false
# pyright: reportUnknownArgumentType=false

# Imports
from __future__ import annotations

import gc
import multiprocessing
import multiprocessing.queues
import time
from collections.abc import Generator, Iterable
from tempfile import TemporaryDirectory
from typing import Any

import mlflow
import numpy as np
from mlflow.entities import Run
from numpy.typing import NDArray
from sklearn.utils import class_weight

from ...decorators import handle_error, measure_time
from ...print import progress, debug, info, warning
from ...ctx import Muffle, MeasureTime
from ...io import clean_path

from .. import mlflow_utils
from ..config.get import DataScienceConfig
from ..dataset import Dataset, DatasetLoader, XyTuple
from ..metric_dictionnary import MetricDictionnary
from ..metric_utils import MetricUtils
from ..utils import Utils
from .abstract_model import AbstractModel

# Constants
MODEL_DOCSTRING: str = """ {model} implementation using advanced model class with common functionality.
For information, refer to the ModelInterface class.
"""
CLASS_ROUTINE_DOCSTRING: str = """ Run the full routine for {model} model.

Args:
	dataset            (Dataset): Dataset to use for training and evaluation.
	kfold              (int): K-fold cross validation index.
	transfer_learning  (str): Pre-trained weights to use, can be "imagenet" or a dataset path like 'data/pizza_not_pizza'.
	verbose            (int): Verbosity level.
	**kwargs           (Any): Additional arguments.

Returns:
	{model}: Trained model instance.
"""

# Base class
class ModelInterface(AbstractModel):
	""" Base class for all models containing common/public methods. """

	# Class constructor
	def __init__(
		self, num_classes: int, kfold: int = 0, transfer_learning: str = "imagenet", **override_params: Any
	) -> None:
		np.random.seed(DataScienceConfig.SEED)
		multiprocessing.set_start_method("spawn", force=True)

		## Base attributes
		self.final_model: Any
		""" Attribute storing the final trained model (Keras model or PyTorch model). """
		self.model_name: str = self.__class__.__name__
		""" Attribute storing the name of the model class, automatically set from the class name.
		Used for logging and display purposes. """
		self.kfold: int = kfold
		""" Attribute storing the number of folds to use for K-fold cross validation.
		If 0 or 1, no K-fold cross validation is used. If > 1, uses K-fold cross validation with that many folds. """
		self.transfer_learning: str = transfer_learning
		""" Attribute storing the transfer learning source, defaults to "imagenet",
		can be set to None or a dataset name present in the data folder. """
		self.is_trained: bool = False
		""" Flag indicating if the model has been trained.
		Must be True before making predictions or evaluating the model. """
		self.num_classes: int = num_classes
		""" Attribute storing the number of classes in the dataset. """
		self.override_params: dict[str, Any] = override_params
		""" Attribute storing the override parameters dictionary for the model. """
		self.run_name: str = ""
		""" Attribute storing the name of the current run, automatically set during training. """
		self.history: list[dict[str, list[float]]] = []
		""" Attribute storing the training history for each fold. """
		self.evaluation_results: list[dict[str, float]] = []
		""" Attribute storing the evaluation results for each fold. """
		self.additional_training_data: XyTuple = XyTuple.empty()
		""" Attribute storing additional training data as a XyTuple
		that is incorporated into the training set right before model fitting.

		This data bypasses cross-validation splitting and is only used during the training phase which
		differs from directly augmenting the dataset via dataset.training_data += additional_training_data,
		which would include the additional data in the cross-validation splitting process.
		"""


		## Model parameters
		# Training parameters
		self.batch_size: int = 8
		""" Attribute storing the batch size for training. """
		self.epochs: int = 50
		""" Attribute storing the number of epochs for training. """
		self.class_weight: dict[int, float] | None = None
		""" Attribute storing the class weights for training, e.g. {0: 0.34, 1: 0.66}. """

		# Fine-tuning parameters
		self.unfreeze_percentage: float = 100
		""" Attribute storing the percentage of layers to fine-tune from the last layer of the base model (0-100). """
		self.fine_tune_last_layers: int = -1
		""" Attribute storing the number of layers to fine-tune, calculated from percentage when total_layers is known. """

		# Optimizer parameters
		self.beta_1: float = 0.95
		""" Attribute storing the beta 1 for Adam optimizer. """
		self.beta_2: float = 0.999
		""" Attribute storing the beta 2 for Adam optimizer. """

		# Callback parameters
		self.early_stop_patience: int = 15
		""" Attribute storing the patience for early stopping. """
		self.model_checkpoint_delay: int = 0
		""" Attribute storing the number of epochs before starting the checkpointing. """

		# ReduceLROnPlateau parameters
		self.learning_rate: float = 1e-4
		""" Attribute storing the learning rate for training. """
		self.reduce_lr_patience: int = 5
		""" Attribute storing the patience for ReduceLROnPlateau. """
		self.min_delta: float = 0.05
		""" Attribute storing the minimum delta for ReduceLROnPlateau (default of the library is 0.0001). """
		self.min_lr: float = 1e-7
		""" Attribute storing the minimum learning rate for ReduceLROnPlateau. """
		self.factor: float = 0.5
		""" Attribute storing the factor for ReduceLROnPlateau. """

		# Warmup parameters
		self.warmup_epochs: int = 5
		""" Attribute storing the number of epochs for learning rate warmup (0 to disable). """
		self.initial_warmup_lr: float = 1e-7
		""" Attribute storing the initial learning rate for warmup. """

		# Learning Rate Finder parameters
		self.lr_finder_min_lr: float = 1e-9
		""" Attribute storing the *minimum* learning rate for the LR Finder. """
		self.lr_finder_max_lr: float = 1.0
		""" Attribute storing the *maximum* learning rate for the LR Finder. """
		self.lr_finder_epochs: int = 3
		""" Attribute storing the number of epochs for the LR Finder. """
		self.lr_finder_update_per_epoch: bool = False
		""" Attribute storing if the LR Finder should increase LR every epoch (True) or batch (False). """
		self.lr_finder_update_interval: int = 5
		""" Attribute storing the number of steps between each lr increase, bigger value means more stable loss. """

		# Unfreeze Percentage Finder parameters
		self.unfreeze_finder_epochs: int = 500
		""" Attribute storing the number of epochs for the Unfreeze Percentage Finder """
		self.unfreeze_finder_update_per_epoch: bool = True
		""" Attribute storing if the Unfreeze Finder should unfreeze every epoch (True) or batch (False). """
		self.unfreeze_finder_update_interval: int = 25
		""" Attribute storing the number of steps between each unfreeze, bigger value means more stable loss. """

		## Model architecture
		self.total_layers: int = 0
		""" Attribute storing the total number of layers in the model. """

	# String representation
	def __str__(self) -> str:
		return f"{self.model_name} (is_trained: {self.is_trained})"

	# Public methods
	@classmethod
	def class_routine(
		cls, dataset: Dataset, kfold: int = 0, transfer_learning: str = "imagenet", verbose: int = 0, **override_params: Any
	) -> ModelInterface:
		return cls(dataset.num_classes, kfold, transfer_learning, **override_params).routine_full(dataset, verbose)

	@measure_time(debug, "Class load (ModelInterface)")
	def class_load(self) -> None:
		""" Clear histories, and set model parameters. """
		# Initialize some attributes
		self.history.clear()
		self.evaluation_results.clear()

		# Get the total number of layers in a subprocess to avoid memory leaks with tensorflow
		with multiprocessing.Pool(1) as pool:
			self.total_layers = pool.apply(self._get_total_layers)

		# Create final model by connecting input to output layer
		self._set_parameters(self.override_params)

	@measure_time(progress)
	@handle_error(error_log=DataScienceConfig.ERROR_LOG)
	def train(self, dataset: Dataset, verbose: int = 0) -> bool:
		""" Method to train the model.

		Args:
			dataset  (Dataset):  Dataset containing the training and testing data.
			verbose  (int):      Level of verbosity, decrease by 1 for each depth
		Returns:
			bool: True if the model was trained successfully.
		Raises:
			ValueError: If the model could not be trained.
		"""
		if not self.class_train(dataset, verbose=verbose):
			raise ValueError("The model could not be trained.")
		self.is_trained = True
		return True


	@handle_error(error_log=DataScienceConfig.ERROR_LOG)
	def predict(self, X_test: Iterable[NDArray[Any]] | Dataset) -> Iterable[NDArray[Any]]:
		""" Method to predict the classes of a batch of data.

		If a Dataset is provided, the test data ungrouped array will be used:
		X_test.test_data.ungrouped_array()[0]

		Otherwise, the input is expected to be an Iterable of NDArray[Any].

		Args:
			X_test  (Iterable[NDArray[Any]] | Dataset):     Features to use for prediction.
		Returns:
			Iterable[NDArray[Any]]: Predictions of the batch.
		Raises:
			ValueError: If the model is not trained.
		"""
		if not self.is_trained:
			raise ValueError("The model must be trained before predicting.")

		# Get X_test from Dataset
		if isinstance(X_test, Dataset):
			return self.class_predict(X_test.test_data.ungrouped_array()[0])
		else:
			return self.class_predict(X_test)


	@measure_time(progress)
	@handle_error(error_log=DataScienceConfig.ERROR_LOG)
	def evaluate(self, dataset: Dataset, verbose: int = 0) -> None:
		""" Method to evaluate the model, it will log metrics and plots to mlflow along with the model.

		Args:
			dataset  (Dataset):  Dataset containing the training and testing data.
			verbose  (int):      Level of verbosity, decrease by 1 for each depth
		"""
		if not self.is_trained:
			raise ValueError("The model must be trained before evaluating.")

		# Metrics (Sensibility, Specificity, AUC, etc.)
		predictions: Iterable[NDArray[Any]] = self.predict(dataset)
		metrics: dict[str, float] = MetricUtils.metrics(dataset, predictions, self.run_name)
		mlflow.log_metrics(metrics)

		# Model specific evaluation
		self.class_evaluate(dataset, save_model=DataScienceConfig.SAVE_MODEL, verbose=verbose)


	@handle_error(error_log=DataScienceConfig.ERROR_LOG)
	@measure_time(progress)
	def routine_full(self, dataset: Dataset, verbose: int = 0) -> ModelInterface:
		""" Method to perform a full routine (load, train and predict, evaluate, and export the model).

		Args:
			dataset  (Dataset):  Dataset containing the training and testing data.
			verbose  (int):      Level of verbosity, decrease by 1 for each depth
		Returns:
			ModelInterface: The model trained and evaluated.
		"""
		# Get the transfer learning weights
		self.transfer_learning = self._get_transfer_learning_weights(dataset, verbose=verbose)

		# Perform the routine
		return self._routine(dataset, verbose=verbose)


	@handle_error(error_log=DataScienceConfig.ERROR_LOG)
	def _routine(self, dataset: Dataset, exp_name: str = "", verbose: int = 0):
		""" Sub-method used in routine_full to perform a full routine

		Args:
			dataset  (Dataset):  Dataset containing the training and testing data.
			exp_name (str):      Name of the experiment (if empty, it will be set automatically)
			verbose  (int):      Level of verbosity, decrease by 1 for each depth
		Returns:
			ModelInterface: The model trained and evaluated.
		"""
		# Init the model
		self.class_load()

		# Start mlflow run silently
		with MeasureTime(message="Experiment setup time"):
			with Muffle(mute_stderr=True):
				exp_name = dataset.get_experiment_name() if exp_name == "" else exp_name
				self.run_name = mlflow_utils.start_run(DataScienceConfig.MLFLOW_URI, exp_name, self.model_name)

		# Log the dataset used for data augmentation and parameters
		if dataset.original_dataset:
			mlflow.log_params({"data_augmentation_based_of": dataset.original_dataset.name})
		self._log_parameters()

		# Train the model
		self.train(dataset, verbose)

		# Evaluate the model
		self.evaluate(dataset, verbose)

		# End mlflow run and return the model
		with Muffle(mute_stderr=True):
			mlflow.end_run()
		return self



	# Protected methods
	def _get_transfer_learning_weights(self, dataset: Dataset, verbose: int = 0) -> str:
		""" Get the transfer learning weights for the model.

		This method handles retrieving pre-trained weights for transfer learning.
		It can:

		1. Return 'imagenet' weights for standard transfer learning
		2. Return None for no transfer learning
		3. Load weights from a previous training run on a different dataset
		4. Train a new model on a different dataset if no previous weights exist

		Returns:
			str: Path to weights file, 'imagenet', or None
		"""
		# If transfer is None or imagenet, return it
		if self.transfer_learning in ("imagenet", None):
			return self.transfer_learning

		# Else, find the weights file path
		else:
			dataset_name: str = clean_path(self.transfer_learning).split("/")[-1]
			exp_name: str = dataset.get_experiment_name(override_name=dataset_name)

			# Find a run with the same model name, and get the weights file
			with Muffle(mute_stderr=True):
				runs: list[Run] = mlflow_utils.get_runs_by_model_name(exp_name, self.model_name)

			# If no runs are found, train a new model on the dataset
			if len(runs) == 0:

				# Load dataset
				pre_dataset: Dataset = DatasetLoader.from_path(
					self.transfer_learning,
					loading_type=dataset.loading_type,
					grouping_strategy=dataset.grouping_strategy
				)
				info(f"In order to do Transfer Learning, training the model on the dataset '{pre_dataset}' first.")

				# Save current settings
				previous_transfer_learning: str = self.transfer_learning
				previous_save_model: bool = DataScienceConfig.SAVE_MODEL
				previous_kfold: int = self.kfold

				# Configure for transfer learning training
				self.transfer_learning = "imagenet"  # Start with imagenet weights
				DataScienceConfig.SAVE_MODEL = True  # Enable model saving
				self.kfold = 0         # Disable k-fold

				# Train model on transfer learning dataset
				self._routine(pre_dataset, exp_name=exp_name, verbose=verbose)

				# Restore previous settings
				self.transfer_learning = previous_transfer_learning
				DataScienceConfig.SAVE_MODEL = previous_save_model
				self.kfold = previous_kfold

				# Get the weights file path - need to refresh the experiment object
				runs: list[Run] = mlflow_utils.get_runs_by_model_name(exp_name, self.model_name)

			# If no runs are found, raise an error
			if not runs:
				raise ValueError(f"No runs found for model {self.model_name} in experiment {exp_name}")
			run: Run = runs[-1]

			# Get the last run's weights path
			# FIXME: Only works if MLFLow URI is file-tree based (not remote or sqlite), which is default
			return mlflow_utils.get_weights_path(from_string=str(run.info.artifact_uri))

	def _get_total_layers(self) -> int:
		""" Get the total number of layers in the model architecture, e.g. 427 for DenseNet121.

		Compatible with Keras/TensorFlow and PyTorch models.

		Returns:
			int: Total number of layers in the model architecture.
		"""
		architecture: Any = self._get_architectures()[1]
		total_layers: int = 0
		# Keras/TensorFlow
		if hasattr(architecture, "layers"):
			total_layers = len(architecture.layers)

		# PyTorch
		elif hasattr(architecture, "children"):
			total_layers = len(architecture.children())

		# Free memory and return the total number of layers
		del architecture
		gc.collect()
		return total_layers

	def _set_parameters(self, override: dict[str, Any] | None = None) -> None:
		""" Set some useful and common models parameters.

		Args:
			override (dict[str, Any]): Dictionary of parameters to override.
		"""
		if override is None:
			override = {}

		# Training parameters
		self.batch_size = override.get("batch_size", self.batch_size)
		self.epochs = override.get("epochs", self.epochs)

		# Callback parameters
		self.early_stop_patience = override.get("early_stop_patience", self.early_stop_patience)
		self.model_checkpoint_delay = override.get("model_checkpoint_delay", self.model_checkpoint_delay)

		# ReduceLROnPlateau parameters
		self.learning_rate = override.get("learning_rate", self.learning_rate)
		self.reduce_lr_patience = override.get("reduce_lr_patience", self.reduce_lr_patience)
		self.min_delta = override.get("min_delta", self.min_delta)
		self.min_lr = override.get("min_lr", self.min_lr)
		self.factor = override.get("factor", self.factor)

		# Warmup parameters
		self.warmup_epochs = override.get("warmup_epochs", self.warmup_epochs)
		self.initial_warmup_lr = override.get("initial_warmup_lr", self.initial_warmup_lr)

		# Fine-tune parameters
		self.unfreeze_percentage = override.get("unfreeze_percentage", self.unfreeze_percentage)
		self.fine_tune_last_layers = max(1, round(self.total_layers * self.unfreeze_percentage / 100))

		# Optimizer parameters
		self.beta_1 = override.get("beta_1", self.beta_1)
		self.beta_2 = override.get("beta_2", self.beta_2)

		# Learning Rate Finder parameters
		self.lr_finder_min_lr = override.get("lr_finder_min_lr", self.lr_finder_min_lr)
		self.lr_finder_max_lr = override.get("lr_finder_max_lr", self.lr_finder_max_lr)
		self.lr_finder_epochs = override.get("lr_finder_epochs", self.lr_finder_epochs)
		self.lr_finder_update_per_epoch = override.get("lr_finder_update_per_epoch", self.lr_finder_update_per_epoch)
		self.lr_finder_update_interval = override.get("lr_finder_update_interval", self.lr_finder_update_interval)

		# Unfreeze Percentage Finder parameters
		self.unfreeze_finder_epochs = override.get("unfreeze_finder_epochs", self.unfreeze_finder_epochs)
		self.unfreeze_finder_update_per_epoch = override.get("unfreeze_finder_update_per_epoch", self.unfreeze_finder_update_per_epoch)
		self.unfreeze_finder_update_interval = override.get("unfreeze_finder_update_interval", self.unfreeze_finder_update_interval)

		# Other parameters
		self.additional_training_data += override.get("additional_training_data", XyTuple.empty())


	def _set_class_weight(self, y_train: NDArray[Any]) -> None:
		""" Calculate class weight for balanced training.

		Args:
			y_train (NDArray[Any]): Training labels
		Returns:
			dict[int, float]: Dictionary mapping class indices to weights, e.g. {0: 0.34, 1: 0.66}
		"""
		# Get the true classes (one-hot -> class indices)
		true_classes: NDArray[Any] = Utils.convert_to_class_indices(y_train)

		# Set the class weights (balanced)
		self.class_weight = dict(enumerate(class_weight.compute_class_weight(
			class_weight="balanced",
			classes=np.unique(true_classes),
			y=true_classes
		)))

	def _log_parameters(self) -> None:
		""" Log the model parameters. """
		mlflow.log_params({
			"cfg_test_size": DataScienceConfig.TEST_SIZE,
			"cfg_validation_size": DataScienceConfig.VALIDATION_SIZE,
			"cfg_seed": DataScienceConfig.SEED,
			"cfg_save_model": DataScienceConfig.SAVE_MODEL,
			"cfg_device": DataScienceConfig.TENSORFLOW_DEVICE,

			# Base attributes
			"param_kfold": self.kfold,
			"param_transfer_learning": self.transfer_learning,

			# Training parameters
			"param_batch_size": self.batch_size,
			"param_epochs": self.epochs,

			# Fine-tuning parameters
			"param_unfreeze_percentage": self.unfreeze_percentage,
			"param_fine_tune_last_layers": self.fine_tune_last_layers,
			"param_total_layers": self.total_layers,

			# Optimizer parameters
			"param_beta_1": self.beta_1,
			"param_beta_2": self.beta_2,
			"param_learning_rate": self.learning_rate,

			# Callback parameters
			"param_early_stop_patience": self.early_stop_patience,
			"param_model_checkpoint_delay": self.model_checkpoint_delay,

			# ReduceLROnPlateau parameters
			"param_reduce_lr_patience": self.reduce_lr_patience,
			"param_min_delta": self.min_delta,
			"param_min_lr": self.min_lr,
			"param_factor": self.factor,

			# Warmup parameters
			"param_warmup_epochs": self.warmup_epochs,
			"param_initial_warmup_lr": self.initial_warmup_lr,
		})

	def _get_fold_split(self, training_data: XyTuple, kfold: int = 5) -> Generator[tuple[XyTuple, XyTuple], None, None]:
		""" Get fold split indices for cross validation.

		This method splits the training data into k folds for cross validation while preserving
		the relationship between original images and their augmented versions.

		The split is done using stratified k-fold to maintain class distribution across folds.
		For each fold, both the training and validation sets contain complete groups of original
		and augmented images.

		Args:
			training_data  (XyTuple): Dataset containing training and test data to split into folds
			kfold          (int):     Number of folds to create
		Returns:
			list[tuple[XyTuple, XyTuple]]: List of (train_data, val_data) tuples for each fold
		"""
		assert kfold not in (0, 1), "kfold must not be 0 or 1"
		yield from training_data.kfold_split(n_splits=kfold, random_state=DataScienceConfig.SEED)

	@handle_error(error_log=DataScienceConfig.ERROR_LOG)
	def _train_final_model(self, dataset: Dataset, verbose: int = 0) -> None:
		""" Train the final model on all data and return it.

		Args:
			dataset  (Dataset):  Dataset containing the training and testing data
			verbose  (int):      Level of verbosity
		"""
		# Get validation data from training data
		debug(f"Training final model on train/val split: {dataset}")

		# Verbose info message
		if verbose > 0:
			info(
				f"({self.model_name}) Training final model on full dataset with "
				f"{len(dataset.training_data.X)} samples ({len(dataset.val_data.X)} validation)"
			)

		# Put the validation data in the test data (since we don't use the test data in the train function)
		old_test_data: XyTuple = dataset.test_data
		dataset.test_data = dataset.val_data

		# Train the final model and remember it
		self.final_model = self._train_fold(dataset, fold_number=0, mlflow_prefix="history_final", verbose=verbose)

		# Restore the old test data
		dataset.test_data = old_test_data
		gc.collect()

	@measure_time(progress)
	@handle_error(error_log=DataScienceConfig.ERROR_LOG)
	def _train_each_fold(self, dataset: Dataset, verbose: int = 0) -> None:
		""" Train the model on each fold and fill self.models with the trained models.

		Args:
			dataset  (Dataset):  Dataset containing the training and testing data
			verbose  (int):      Level of verbosity
		"""
		# Get fold split
		fold_split: Generator[tuple[XyTuple, XyTuple], None, None] = self._get_fold_split(dataset.training_data, self.kfold)

		# Train on each fold
		for i, (train_data, test_data) in enumerate(fold_split):
			fold_number: int = i + 1

			# During Cross Validation, the validation data is the same as the test data.
			# Except when the validation population is 1 sample (e.g. LeaveOneOut)
			# Therefore, we need to use the original validation data for the final model
			if self.kfold < 0 or len(test_data.X) == 1:
				val_data: XyTuple = dataset.val_data
			else:
				val_data: XyTuple = test_data

			# Create a new dataset (train/val based of training data)
			new_dataset: Dataset = Dataset(
				training_data=train_data,
				val_data=val_data,
				test_data=test_data,
				name=dataset.name,
				grouping_strategy=dataset.grouping_strategy,
				labels=dataset.labels
			)

			# Log the fold
			if verbose > 0:
				# If there are multiple validation samples or no filepaths, show the number of validation samples
				if len(test_data.X) != 1 or not test_data.filepaths:
					debug(
						f"({self.model_name}) Fold {fold_number} training with "
						f"{len(train_data.X)} samples ({len(test_data.X)} validation)"
					)
				# Else, show the filepath of the single validation sample (useful for debugging)
				else:
					debug(
						f"({self.model_name}) Fold {fold_number} training with "
						f"{len(train_data.X)} samples (validation: {test_data.filepaths[0]})"
					)

			# Train the model on the fold
			handle_error(self._train_fold,
				message=f"({self.model_name}) Fold {fold_number} training failed", error_log=DataScienceConfig.ERROR_LOG
			)(
				dataset=new_dataset,
				fold_number=fold_number,
				mlflow_prefix=f"history_fold_{fold_number}",
				verbose=verbose
			)

			# Collect garbage to free up some memory
			gc.collect()

	@measure_time(progress)
	@handle_error(error_log=DataScienceConfig.ERROR_LOG)
	def class_train(self, dataset: Dataset, verbose: int = 0) -> bool:
		""" Train the model using k-fold cross validation and then full model training.

		Args:
			dataset (Dataset):  Dataset containing the training and testing data (test data will not be used in class_train)
			verbose (int):      Level of verbosity
		Returns:
			bool: True if training was successful
		"""
		# Compute the class weights
		self._set_class_weight(np.array(dataset.training_data.y))

		# Find the best learning rate
		if DataScienceConfig.DO_LEARNING_RATE_FINDER > 0:
			info(f"({self.model_name}) Finding the best learning rate...")
			found_lr: float | None = self._find_best_learning_rate(dataset, verbose)
			if DataScienceConfig.DO_LEARNING_RATE_FINDER == 2 and found_lr is not None:
				self.learning_rate = found_lr
				mlflow.log_params({"param_learning_rate": found_lr})
				info(f"({self.model_name}) Now using learning rate: {found_lr:.2e}")

		# Find the best unfreeze percentage
		if DataScienceConfig.DO_UNFREEZE_FINDER > 0:
			info(f"({self.model_name}) Finding the best unfreeze percentage...")
			found_unfreeze: float | None = self._find_best_unfreeze_percentage(dataset, verbose)
			if DataScienceConfig.DO_UNFREEZE_FINDER == 2 and found_unfreeze is not None:
				self.unfreeze_percentage = found_unfreeze
				info(f"({self.model_name}) Now using unfreeze percentage: {found_unfreeze:.2f}%")

		# If k-fold is enabled, train the model on each fold
		if self.kfold not in (0, 1):
			self._train_each_fold(dataset, verbose)

		# Train the final model on all data
		self._train_final_model(dataset, verbose)
		return True

	@handle_error(error_log=DataScienceConfig.ERROR_LOG)
	def _log_metrics(self, from_index: int = 1) -> None:
		""" Calculate (average and standard deviation) and log metrics for each evaluation result.

		Args:
			from_index (int): Index of the first evaluation result to use
		"""
		# For each metric, calculate the average and standard deviation
		for metric_name in self.evaluation_results[0].keys():

			# Get the metric values for each fold
			metric_values: list[float] = [x[metric_name] for x in self.evaluation_results[from_index:]]
			if not metric_values:
				continue

			# Log the average and standard deviation
			avg_key: str = MetricDictionnary.AVERAGE_METRIC.replace("METRIC_NAME", metric_name)
			std_key: str = MetricDictionnary.STANDARD_DEVIATION_METRIC.replace("METRIC_NAME", metric_name)
			mlflow.log_metric(avg_key, float(np.mean(metric_values)))
			mlflow.log_metric(std_key, float(np.std(metric_values)))

	@handle_error(error_log=DataScienceConfig.ERROR_LOG)
	def class_evaluate(
		self,
		dataset: Dataset,
		metrics_names: tuple[str, ...] = (),
		save_model: bool = False,
		verbose: int = 0
	) -> bool:
		""" Evaluate the model using the given predictions and labels.

		Args:
			dataset         (Dataset):   Dataset containing the training and testing data
			metrics_names   (list[str]): List of metrics to plot (default to all metrics)
			save_model      (bool):      Whether to save the best model
			verbose         (int):       Level of verbosity
		Returns:
			bool: True if evaluation was successful
		"""
		# If no metrics names are provided, use all metrics
		if not metrics_names:
			metrics_names = tuple(self.evaluation_results[0].keys())

		# Log metrics and plot curves
		MetricUtils.plot_every_metric_curves(self.history, metrics_names, self.run_name)
		self._log_metrics()

		# Save the best model if save_model is True
		if save_model:
			if verbose > 0:
				with MeasureTime(debug, "Saving best model"):
					self._log_final_model()
			else:
				self._log_final_model()

		# Success
		return True


	# Protected methods for training
	@handle_error(error_log=DataScienceConfig.ERROR_LOG)
	def _find_best_learning_rate(self, dataset: Dataset, verbose: int = 0) -> float:
		""" Find the best learning rate for the model, optionally using a subprocess.

		Args:
			dataset  (Dataset):  Dataset to use for training.
			verbose  (int):      Verbosity level (controls progress bar).
		Returns:
			float: The best learning rate found.
		"""
		results: dict[str, Any] = {}
		for try_count in range(10):
			try:
				if DataScienceConfig.DO_FIT_IN_SUBPROCESS:
					queue: multiprocessing.queues.Queue[dict[str, Any]] = multiprocessing.Queue()
					process: multiprocessing.Process = multiprocessing.Process(
						target=self._find_best_learning_rate_subprocess,
						kwargs={"dataset": dataset, "queue": queue, "verbose": verbose}
					)
					process.start()
					process.join()
					results = queue.get(timeout=60)
				else:
					results = self._find_best_learning_rate_subprocess(dataset, verbose=verbose)
				if results:
					break
			except Exception as e:
				warning(f"Error finding best learning rate: {e}\nRetrying in 60 seconds ({try_count + 1}/10)...")
				time.sleep(60)

		# Plot the learning rate vs loss and find the best learning rate
		return MetricUtils.find_best_x_and_plot(
			results["learning_rates"],
			results["losses"],
			smoothen=True,
			use_steep=True,
			run_name=self.run_name,
			x_label="Learning Rate",
			y_label="Loss",
			plot_title="Learning Rate Finder",
			log_x=True,
			y_limits=(0, 4.0)
		)

	@handle_error(error_log=DataScienceConfig.ERROR_LOG)
	def _find_best_unfreeze_percentage(self, dataset: Dataset, verbose: int = 0) -> float:
		""" Find the best unfreeze percentage for the model, optionally using a subprocess.

		Args:
			dataset  (Dataset):  Dataset to use for training.
			verbose  (int):      Verbosity level (controls progress bar).
		Returns:
			float: The best unfreeze percentage found.
		"""
		results: dict[str, Any] = {}
		for try_count in range(10):
			try:
				if DataScienceConfig.DO_FIT_IN_SUBPROCESS:
					queue: multiprocessing.queues.Queue[dict[str, Any]] = multiprocessing.Queue()
					process: multiprocessing.Process = multiprocessing.Process(
						target=self._find_best_unfreeze_percentage_subprocess,
						kwargs={"dataset": dataset, "queue": queue, "verbose": verbose}
					)
					process.start()
					process.join()
					results = queue.get(timeout=60)
				else:
					results = self._find_best_unfreeze_percentage_subprocess(dataset, verbose=verbose)
				if results:
					break
			except Exception as e:
				warning(f"Error finding best unfreeze percentage: {e}\nRetrying in 60 seconds ({try_count + 1}/10)...")
				time.sleep(60)

		# Plot the unfreeze percentage vs loss and find the best unfreeze percentage
		return MetricUtils.find_best_x_and_plot(
			results["unfreeze_percentages"],
			results["losses"],
			smoothen=True,
			use_steep=False,
			run_name=self.run_name,
			x_label="Unfreeze Percentage",
			y_label="Loss",
			plot_title="Unfreeze Percentage Finder",
			log_x=False,
			y_limits=(0, 4.0)
		)


	@measure_time(progress)
	def _train_fold(self, dataset: Dataset, fold_number: int = 0, mlflow_prefix: str = "history", verbose: int = 0) -> Any:
		""" Train model on a single fold.

		Args:
			dataset      (Dataset):  Dataset to train on
			fold_number  (int):      Fold number (0 for final model)
			prefix       (str):      Prefix for the history
			verbose      (int):      Verbosity level
		"""
		# Create the checkpoint path
		checkpoint_path: str = f"{DataScienceConfig.TEMP_FOLDER}/{self.run_name}_best_model_fold_{fold_number}.keras"

		# Prepare visualization arguments if needed
		temp_dir: TemporaryDirectory[str] | None = None
		if DataScienceConfig.DO_SALIENCY_AND_GRADCAM and dataset.test_data.n_samples == 1:
			temp_dir = TemporaryDirectory()

		# Create and run the process
		return_values: dict[str, Any] = {}
		for try_count in range(10):
			try:
				if DataScienceConfig.DO_FIT_IN_SUBPROCESS and fold_number > 0:
					queue: multiprocessing.queues.Queue[dict[str, Any]] = multiprocessing.Queue()
					process: multiprocessing.Process = multiprocessing.Process(
						target=self._train_subprocess,
						args=(dataset, checkpoint_path, temp_dir),
						kwargs={"queue": queue, "verbose": verbose}
					)
					process.start()
					process.join()
					return_values = queue.get(timeout=60)
				else:
					return_values = self._train_subprocess(dataset, checkpoint_path, temp_dir, verbose=verbose)
				if return_values:
					break
			except Exception as e:
				warning(f"Error during _train_fold: {e}\nRetrying in 60 seconds ({try_count + 1}/10)...")
				time.sleep(60)
		history: dict[str, Any] = return_values["history"]
		eval_results: dict[str, Any] = return_values["eval_results"]
		predictions: NDArray[Any] = return_values["predictions"]
		true_classes: NDArray[Any] = return_values["true_classes"]
		training_predictions: NDArray[Any] = return_values.get("training_predictions", None)
		training_true_classes: NDArray[Any] = return_values.get("training_true_classes", None)

		# For each epoch, log the history
		mlflow_utils.log_history(history, prefix=mlflow_prefix)

		# Append the history and evaluation results
		self.history.append(history)
		self.evaluation_results.append(eval_results)

		# Generate and save ROC Curve and PR Curve for this fold
		MetricUtils.all_curves(true_classes, predictions, fold_number, run_name=self.run_name)

		# If final model, also log the ROC curve and PR curve for the train set
		if fold_number == 0:
			fold_number = -2	# -2 is the train set
			MetricUtils.all_curves(training_true_classes, training_predictions, fold_number, run_name=self.run_name)

		# Log visualization artifacts if they were generated
		if temp_dir is not None:
			mlflow.log_artifacts(temp_dir.name)
			temp_dir.cleanup()

		# Show some metrics
		if verbose > 0:
			last_history: dict[str, Any] = {k: v[-1] for k, v in history.items()}
			info(f"Training done, metrics: {last_history}")

		# Return the trained model
		return return_values.get("model", None)




"""
This module provides utility functions for printing messages with different levels of importance.

If a message is printed multiple times, it will be displayed as "(xN) message"
where N is the number of times the message has been printed.

.. image:: https://raw.githubusercontent.com/Stoupy51/stouputils/refs/heads/main/assets/print_module.gif
  :alt: stouputils print examples
"""

# Imports
import os
import sys
import time
from collections.abc import Callable, Iterable, Iterator
from typing import IO, Any, TextIO, TypeVar, cast

from tqdm.auto import tqdm

# Colors constants
RESET: str   = "\033[0m"
RED: str     = "\033[91m"
GREEN: str   = "\033[92m"
YELLOW: str  = "\033[93m"
BLUE: str    = "\033[94m"
MAGENTA: str = "\033[95m"
CYAN: str    = "\033[96m"
LINE_UP: str = "\033[1A"

# Constants
BAR_FORMAT: str = "{l_bar}{bar}" + MAGENTA + "| {n_fmt}/{total_fmt} [{rate_fmt}{postfix}, {elapsed}<{remaining}]" + RESET
T = TypeVar("T")

# Enable colors on Windows 10 terminal if applicable
if os.name == "nt":
	os.system("color")

# Print functions
previous_args_kwards: tuple[Any, Any] = ((), {})
nb_values: int = 1
import_time: float = time.time()

# Colored for loop function
def colored_for_loop(
	iterable: Iterable[T],
	desc: str = "Processing",
	color: str = MAGENTA,
	bar_format: str = BAR_FORMAT,
	ascii: bool = False,
	**kwargs: Any
) -> Iterator[T]:
	""" Function to iterate over a list with a colored TQDM progress bar like the other functions in this module.

	Args:
		iterable	(Iterable):			List to iterate over
		desc		(str):				Description of the function execution displayed in the progress bar
		color		(str):				Color of the progress bar (Defaults to MAGENTA)
		bar_format	(str):				Format of the progress bar (Defaults to BAR_FORMAT)
		ascii		(bool):				Whether to use ASCII or Unicode characters for the progress bar (Defaults to False)
		verbose		(int):				Level of verbosity, decrease by 1 for each depth (Defaults to 1)
		**kwargs:						Additional arguments to pass to the TQDM progress bar

	Yields:
		T:		Each item of the iterable

	Examples:
		>>> for i in colored_for_loop(range(10), desc="Time sleeping loop"):
		...     time.sleep(0.01)
		>>> # Time sleeping loop: 100%|██████████████████| 10/10 [ 95.72it/s, 00:00<00:00]
	"""
	if bar_format == BAR_FORMAT:
		bar_format = bar_format.replace(MAGENTA, color)
	desc = color + desc

	yield from tqdm(iterable, desc=desc, bar_format=bar_format, ascii=ascii, **kwargs)

def info(
	*values: Any,
	color: str = GREEN,
	text: str = "INFO ",
	prefix: str = "",
	file: TextIO | list[TextIO] | None = None,
	**print_kwargs: Any,
) -> None:
	""" Print an information message looking like "[INFO HH:MM:SS] message" in green by default.

	Args:
		values			(Any):					Values to print (like the print function)
		color			(str):					Color of the message (default: GREEN)
		text			(str):					Text of the message (default: "INFO ")
		prefix			(str):					Prefix to add to the values
		file			(TextIO|list[TextIO]):	File(s) to write the message to (default: sys.stdout)
		print_kwargs	(dict):					Keyword arguments to pass to the print function
	"""
	# Use stdout if no file is specified
	if file is None:
		file = sys.stdout

	# If file is a list, recursively call info() for each file
	if isinstance(file, list):
		for f in file:
			info(*values, color=color, text=text, prefix=prefix, file=f, **print_kwargs)
	else:
		# Build the message with prefix, color, text and timestamp
		message: str = f"{prefix}{color}[{text} {current_time()}]"

		# If this is a repeated print, add a line up and counter
		if is_same_print(*values, color=color, text=text, prefix=prefix, **print_kwargs):
			message = f"{LINE_UP}{message} (x{nb_values})"

		# Print the message with the values and reset color
		print(message, *values, RESET, file=file, **print_kwargs)

def debug(*values: Any, **print_kwargs: Any) -> None:
	""" Print a debug message looking like "[DEBUG HH:MM:SS] message" in cyan by default. """
	if "text" not in print_kwargs:
		print_kwargs["text"] = "DEBUG"
	if "color" not in print_kwargs:
		print_kwargs["color"] = CYAN
	info(*values, **print_kwargs)

def alt_debug(*values: Any, **print_kwargs: Any) -> None:
	""" Print a debug message looking like "[DEBUG HH:MM:SS] message" in blue by default. """
	if "text" not in print_kwargs:
		print_kwargs["text"] = "DEBUG"
	if "color" not in print_kwargs:
		print_kwargs["color"] = BLUE
	info(*values, **print_kwargs)

def suggestion(*values: Any, **print_kwargs: Any) -> None:
	""" Print a suggestion message looking like "[SUGGESTION HH:MM:SS] message" in cyan by default. """
	if "text" not in print_kwargs:
		print_kwargs["text"] = "SUGGESTION"
	if "color" not in print_kwargs:
		print_kwargs["color"] = CYAN
	info(*values, **print_kwargs)

def progress(*values: Any, **print_kwargs: Any) -> None:
	""" Print a progress message looking like "[PROGRESS HH:MM:SS] message" in magenta by default. """
	if "text" not in print_kwargs:
		print_kwargs["text"] = "PROGRESS"
	if "color" not in print_kwargs:
		print_kwargs["color"] = MAGENTA
	info(*values, **print_kwargs)

def warning(*values: Any, **print_kwargs: Any) -> None:
	""" Print a warning message looking like "[WARNING HH:MM:SS] message" in yellow by default and in sys.stderr. """
	if "file" not in print_kwargs:
		print_kwargs["file"] = sys.stderr
	if "text" not in print_kwargs:
		print_kwargs["text"] = "WARNING"
	if "color" not in print_kwargs:
		print_kwargs["color"] = YELLOW
	info(*values, **print_kwargs)

def error(*values: Any, exit: bool = False, **print_kwargs: Any) -> None:
	""" Print an error message (in sys.stderr and in red by default)
	and optionally ask the user to continue or stop the program.

	Args:
		values			(Any):		Values to print (like the print function)
		exit			(bool):		Whether to ask the user to continue or stop the program,
			false to ignore the error automatically and continue
		print_kwargs	(dict):		Keyword arguments to pass to the print function
	"""
	file: TextIO = sys.stderr
	if "file" in print_kwargs:
		if isinstance(print_kwargs["file"], list):
			file = cast(TextIO, print_kwargs["file"][0])
		else:
			file = print_kwargs["file"]
	if "text" not in print_kwargs:
		print_kwargs["text"] = "ERROR"
	if "color" not in print_kwargs:
		print_kwargs["color"] = RED
	info(*values, **print_kwargs)
	if exit:
		try:
			print("Press enter to ignore error and continue, or 'CTRL+C' to stop the program... ", file=file)
			input()
		except (KeyboardInterrupt, EOFError):
			print(file=file)
			sys.exit(1)

def whatisit(
	*values: Any,
	print_function: Callable[..., None] = debug,
	max_length: int = 250,
	color: str = CYAN,
	**print_kwargs: Any,
) -> None:
	""" Print the type of each value and the value itself, with its id and length/shape.

	The output format is: "type, <id id_number>:	(length/shape) value"

	Args:
		values			(Any):		Values to print
		print_function	(Callable):	Function to use to print the values (default: debug())
		max_length		(int):		Maximum length of the value string to print (default: 250)
		color			(str):		Color of the message (default: CYAN)
		print_kwargs	(dict):		Keyword arguments to pass to the print function
	"""
	def _internal(value: Any) -> str:
		""" Get the string representation of the value, with length or shape instead of length if shape is available """

		# Get the length or shape of the value
		length: str = ""
		try:
			length = f"(length: {len(value)}) "
		except (AttributeError, TypeError):
			pass
		try:
			length = f"(shape: {value.shape}) "
		except (AttributeError, TypeError):
			pass

		# Get the dtype if available
		dtype: str = ""
		try:
			dtype = f"(dtype: {value.dtype}) "
		except (AttributeError, TypeError):
			pass

		# Get the string representation of the value
		value_str: str = str(value)
		if len(value_str) > max_length:
			value_str = value_str[:max_length] + "..."
		if "\n" in value_str:
			value_str = "\n" + value_str	# Add a newline before the value if there is a newline in it.

		# Return the formatted string
		return f"{type(value)}, <id {id(value)}>: {dtype}{length}{value_str}"

	# Add the color to the message
	if "color" not in print_kwargs:
		print_kwargs["color"] = color

	# Print the values
	if len(values) > 1:
		print_function("(What is it?)", **print_kwargs)
		for value in values:
			print_function(_internal(value), **print_kwargs)
	elif len(values) == 1:
		print_function(f"(What is it?) {_internal(values[0])}", **print_kwargs)

def breakpoint(*values: Any, print_function: Callable[..., None] = warning, **print_kwargs: Any) -> None:
	""" Breakpoint function, pause the program and print the values.

	Args:
		values			(Any):		Values to print
		print_function	(Callable):	Function to use to print the values (default: warning())
		print_kwargs	(dict):		Keyword arguments to pass to the print function
	"""
	if "text" not in print_kwargs:
		print_kwargs["text"] = "BREAKPOINT (press Enter)"
	file: TextIO = sys.stderr
	if "file" in print_kwargs:
		if isinstance(print_kwargs["file"], list):
			file = cast(TextIO, print_kwargs["file"][0])
		else:
			file = print_kwargs["file"]
	whatisit(*values, print_function=print_function, **print_kwargs)
	try:
		input()
	except (KeyboardInterrupt, EOFError):
		print(file=file)
		sys.exit(1)


# TeeMultiOutput class to duplicate output to multiple file-like objects
class TeeMultiOutput:
	""" File-like object that duplicates output to multiple file-like objects.

	Args:
		*files         (IO[Any]):  One or more file-like objects that have write and flush methods
		strip_colors   (bool):     Strip ANSI color codes from output sent to non-stdout/stderr files
		ascii_only     (bool):     Replace non-ASCII characters with their ASCII equivalents for non-stdout/stderr files
		ignore_lineup  (bool):     Ignore lines containing LINE_UP escape sequence in non-terminal outputs

	Examples:
		>>> f = open("logfile.txt", "w")
		>>> sys.stdout = TeeMultiOutput(sys.stdout, f)
		>>> print("Hello World")  # Output goes to both console and file
		Hello World
		>>> f.close()	# TeeMultiOutput will handle any future writes to closed files gracefully
	"""
	def __init__(
		self, *files: IO[Any], strip_colors: bool = True, ascii_only: bool = True, ignore_lineup: bool = True
	) -> None:
		# Flatten any TeeMultiOutput instances in files
		flattened_files: list[IO[Any]] = []
		for file in files:
			if isinstance(file, TeeMultiOutput):
				flattened_files.extend(file.files)
			else:
				flattened_files.append(file)

		self.files: tuple[IO[Any], ...] = tuple(flattened_files)
		""" File-like objects to write to """
		self.strip_colors: bool = strip_colors
		""" Whether to strip ANSI color codes from output sent to non-stdout/stderr files """
		self.ascii_only: bool = ascii_only
		""" Whether to replace non-ASCII characters with their ASCII equivalents for non-stdout/stderr files """
		self.ignore_lineup: bool = ignore_lineup
		""" Whether to ignore lines containing LINE_UP escape sequence in non-terminal outputs """

	@property
	def encoding(self) -> str:
		""" Get the encoding of the first file, or "utf-8" as fallback.

		Returns:
			str: The encoding, ex: "utf-8", "ascii", "latin1", etc.
		"""
		try:
			return self.files[0].encoding	# type: ignore
		except (IndexError, AttributeError):
			return "utf-8"

	def write(self, obj: str) -> int:
		""" Write the object to all files while stripping colors if needed.

		Args:
			obj (str): String to write
		Returns:
			int: Number of characters written to the first file
		"""
		files_to_remove: list[IO[Any]] = []
		num_chars_written: int = 0
		for i, f in enumerate(self.files):
			try:
				# Check if file is closed
				if hasattr(f, "closed") and f.closed:
					files_to_remove.append(f)
					continue

				# Check if this file is a terminal/console or a regular file
				content: str = obj
				if not (hasattr(f, "isatty") and f.isatty()):
					# Non-terminal files get processed content (stripped colors, ASCII-only, etc.)

					# Skip content if it contains LINE_UP and ignore_lineup is True
					if self.ignore_lineup and (LINE_UP in content or "\r" in content):
						continue

					# Strip colors if needed
					if self.strip_colors:
						content = remove_colors(content)

					# Replace Unicode block characters with ASCII equivalents
					# Replace other problematic Unicode characters as needed
					if self.ascii_only:
						content = content.replace('█', '#')
						content = ''.join(c if ord(c) < 128 else '?' for c in content)

				# Write content to file
				if i == 0:
					num_chars_written = f.write(content)
				else:
					f.write(content)

			except ValueError:
				# ValueError is raised when writing to a closed file
				files_to_remove.append(f)
			except Exception:
				pass

		# Remove closed files from the list
		if files_to_remove:
			self.files = tuple(f for f in self.files if f not in files_to_remove)
		return num_chars_written

	def flush(self) -> None:
		""" Flush all files. """
		for f in self.files:
			try:
				f.flush()
			except Exception:
				pass

	def fileno(self) -> int:
		""" Return the file descriptor of the first file. """
		return self.files[0].fileno() if hasattr(self.files[0], "fileno") else 0


# Utility functions
def remove_colors(text: str) -> str:
	""" Remove the colors from a text """
	for color in [RESET, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, LINE_UP]:
		text = text.replace(color, "")
	return text

def is_same_print(*args: Any, **kwargs: Any) -> bool:
	""" Checks if the current print call is the same as the previous one. """
	global previous_args_kwards, nb_values
	if previous_args_kwards == (args, kwargs):
		nb_values += 1
		return True
	else:
		previous_args_kwards = (args, kwargs)
		nb_values = 1
		return False

def current_time() -> str:
	""" Get the current time as "HH:MM:SS" if less than 24 hours since import, else "YYYY-MM-DD HH:MM:SS" """
	# If the import time is more than 24 hours, return the full datetime
	if (time.time() - import_time) > (24 * 60 * 60):
		return time.strftime("%Y-%m-%d %H:%M:%S")
	else:
		return time.strftime("%H:%M:%S")


def show_version(main_package: str = "stouputils", primary_color: str = CYAN, secondary_color: str = GREEN):
	""" Print the version of the main package and its dependencies.

	Used by the "stouputils --version" command.

	Args:
		main_package			(str):			Name of the main package to show version for
		primary_color			(str):			Color to use for the primary package name
		secondary_color			(str):			Color to use for the secondary package names
	"""
	# Imports
	from importlib.metadata import requires, version
	def ver(package_name: str) -> str:
		try:
			return version(package_name)
		except Exception:
			return ""

	# Get dependencies dynamically and extract package names from requirements (e.g., "tqdm>=4.0.0" -> "tqdm")
	deps: list[str] = requires(main_package) or []
	dep_names: list[str] = sorted([
		dep
			.split(">")[0]
			.split("<")[0]
			.split("=")[0]
			.split("[")[0]
		for dep in deps
	])
	all_deps: list[tuple[str, str]] = [
		(x, ver(x).split("version: ")[-1])
		for x in (main_package, *dep_names)
	]
	all_deps = [pair for pair in all_deps if pair[1]]  # Filter out packages with no version found
	longest_name_length: int = max(len(name) for name, _ in all_deps)
	longest_version_length: int = max(len(ver) for _, ver in all_deps)

	for pkg, v in all_deps:
		pkg_spacing: str = " " * (longest_name_length - len(pkg))

		# Highlight the main package with a different style
		if pkg == main_package:
			separator: str = "─" * (longest_name_length + longest_version_length + 4)
			print(f"{primary_color}{separator}{RESET}")
			print(f"{primary_color}{pkg}{pkg_spacing}  {secondary_color}v{v}{RESET}")
			print(f"{primary_color}{separator}{RESET}")
		else:
			print(f"{primary_color}{pkg}{pkg_spacing}  {secondary_color}v{v}{RESET}")
	return

# Test the print functions
if __name__ == "__main__":
	info("Hello", "World")
	time.sleep(0.5)
	info("Hello", "World")
	time.sleep(0.5)
	info("Hello", "World")
	time.sleep(0.5)
	info("Not Hello World !")
	time.sleep(0.5)
	info("Hello", "World")
	time.sleep(0.5)
	info("Hello", "World")

	# All remaining print functions
	debug("Hello", "World")
	suggestion("Hello", "World")
	progress("Hello", "World")
	warning("Hello", "World")
	error("Hello", "World", exit=False)
	whatisit("Hello", "World")

